#include <stdio.h>
#include <sys/poll.h> /* poll() */
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/uio.h> /* writev() */
#include <unistd.h>
#include <netdb.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#include <signal.h>
#include <fcntl.h> /* logging */
#include "config.h"

/* defines */
#define MAXLEN		512			/* Max length of a line.
						   510 chars + CR + LF */
#define HOSTLEN		64			/* Max length of a FQDN */
#define USERLEN		10			/* Max length of a username */
#define NICKLEN		15			/* EFnet uses 9 */
#define VERLEN		128			/* Length of ircd version */
/* Receive buffer size.
 * On small nets or nets using STOCK 2.8.21 use 8196.
 * On larger nets running +CS, hybrid, or ratbox servers, use 65536.
 * NOTE: On hybrid servers use /stats ? to see if you're going over 8kb/s. */
#define BUFSIZE         16384
#define TS_CURRENT	5
#define TS_MIN		3
#define AS_VER		"AlcStats_3.5"
#define DAT_VER		3

/* SVR4&SUS use IOV_MAX, 4.3+BSD uses UIO_MAXIOV.
 * Some FreeBSDs don't have IOV_MAX and UIO_MAXIOV is KERNEL-only, so we'll go
 * ahead and define it to 1024, which is what they support anyways. */
#ifndef IOV_MAX
# ifdef UIO_MAXIOV
# define IOV_MAX UIO_MAXIOV
# else
# define IOV_MAX 1024
# endif
#endif

/* globals */
#define LOGSIZE 1024	/* Size of logging buffer */
extern struct opt		opts;
extern struct confitems		conf;
extern struct pollfd		pollfd[1];
extern struct total_struct	totals;
extern pid_t			pid;
extern time_t			start, now, nextc;
extern char			interface[USERLEN + HOSTLEN + 2];
extern char			logbuf[LOGSIZE+1];
extern int			logfd;
extern time_t			now;
extern char			connected;
extern int			stop;
extern unsigned int		iovs;

/* structures */
struct opt
{
	char		*conffile;
	char		*logfile;
	unsigned char	qs : 1;
	unsigned char	debug : 1;
	unsigned char	fork : 1;
};

struct confitems {
	char    *admin;
	char    *datafile;
	char    *location;
	char    *network;
	char    *rrddir;
	char    *servname;
	char    *servdesc;
	char    *username;
	int	calctime;
	int	domainexpire;
	int	domainheap;
	int     domainhsize;
	int	serverexpire;
	int	serverheap;
	int	serverhsize;
	int	userheap;
	int     userhsize;
	int     maxthresh;
};

struct total_struct
{
	unsigned int	checks;
	unsigned long	tot_users;
	unsigned long	tot_opers;
	unsigned long	tot_dupes;
	unsigned int	cur_users;
	unsigned int	cur_opers;
	unsigned int	cur_dupes;
	unsigned int	max_users;
	unsigned int	max_opers;
	unsigned int	max_dupes;
	unsigned short	cur_servers;
	unsigned short	max_servers;
	unsigned short	tot_servers;
	unsigned short	domains;
	unsigned short	splits;
	unsigned int	skills;
	unsigned int	okills;
};

/* prototypes */
extern void *MyMalloc(size_t);
extern void MyFree(void *);
extern char *mystrdup(char *);
extern void parse_cmdline(int, char **);
extern char *mystrlcpy(char *, const char *, size_t);
extern char *host_to_domain(const char *, char *);
extern void init_totals(void);
extern void stats_msg(int, char **);
extern void stats_missing(int, char **);
extern void stats_help(int, char **);
extern void stats_net(int, char **);
extern void stats_stats(int, char **);
extern void stats_map(int, char **);
extern void stats_delete(int, char **);
extern void read_data(void);
extern void write_data(void);
extern void check_old(void);
extern void clear_opts(void);
extern void downstring(char *);
extern void die(const char *, ...);
extern void sigterm_handler(int);
