/*
 *  ircd-ratbox: A slightly useful ircd.
 *  balloc.h: The ircd block allocator header.
 *
 *  Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 *  Copyright (C) 1996-2002 Hybrid Development Team
 *  Copyright (C) 2002-2004 ircd-ratbox development team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: balloc.h,v 1.2 2004/07/11 01:04:13 jailbird Exp $
 */

#ifndef NOBALLOC

/* 
 * Block contains status information for an allocated block in our
 * heap.
 */

struct Block {
	int			freeElems;	/* Number of available elems */
	size_t			alloc_size;
	struct Block		*next;		/* Next in our chain of blocks */
	void			*elems;		/* Points to allocated memory */
	struct list_struct	*free_list;
	struct list_struct	*used_list;                                      
};
typedef struct Block Block;

struct MemBlock
{
	struct list_struct	self;
	Block			*block;	/* Which block we belong to */
};
typedef struct MemBlock MemBlock;

/* 
 * BlockHeap contains the information for the root node of the
 * memory heap.
 */
struct BlockHeap {
	size_t			elemSize;		/* Size of each element to be stored */
	int			elemsPerBlock;		/* Number of elements per block */
	int			blocksAllocated;	/* Number of blocks allocated */
	int			freeElems;		/* Number of free elements */
	Block			*base;			/* Pointer to first block */
	struct BlockHeap	*next;			/* Pointer to next heap */
};
typedef struct BlockHeap BlockHeap;


extern int		BlockHeapFree(BlockHeap *, void *);
extern void		*BlockHeapAlloc(BlockHeap *);
extern BlockHeap	*BlockHeapCreate(size_t, int);
extern int		BlockHeapDestroy(BlockHeap *);
extern void		initBlockHeap(void);
extern void		heap_garbage_collection(void);

#endif

