#include "alcstats.h"
#include "serv.h"
#include "list.h"
#include "balloc.h"
#include "hash.h"
#include "servlist.h"
#include "userlist.h"
#include "domainlist.h"
#ifdef USE_RRD
#include "rrdtool.h"
#endif

#define USER_COL	50

time_t			nextc;
static char		buf[MAXLEN+1];
static struct iovec	iov[IOV_MAX];
unsigned int		iovs;

void dump_map(char *nickname, SERVPTR root, char *pbuf)
{
	int	len, cnt=0, i=0;
	LISTPTR	ptr;
	SERVPTR	server;

	*pbuf = '\0';
	strncat(pbuf, root->servername, MAXLEN-((size_t)pbuf-(size_t)buf));
	len = strlen(buf);
	buf[len] = ' ';

	if (len < USER_COL) {
		for (i = len+1; i < USER_COL; ++i) {
			buf[i] = '-';
		}
	}

	snprintf(buf+USER_COL, MAXLEN-USER_COL, " | Users: %5u (%4.1f%%)", root->cur_users,
		 100.0*(double)root->cur_users/(double)totals.cur_users);
	writeln(":%s NOTICE %s :%s", conf.servname, nickname, buf);

	if (root->servers != NULL) {
		cnt += root->links;
		if (cnt) {
			if (pbuf > (buf+3)) {
				pbuf[-2] = ' ';
				if (pbuf[-3] == '`')
					pbuf[-3] = ' ';
			}
		}
	}

	i = 1;
	for (ptr = root->servers; ptr != NULL; ptr = ptr->next) {
		server = ptr->data;
		*pbuf = ' ';
		if (i < cnt)
			*(pbuf + 1) = '|';
		else
			*(pbuf + 1) = '`';
		*(pbuf + 2) = '-';
		*(pbuf + 3) = ' ';
		dump_map(nickname, server, pbuf+4);
		++i;
	}
}

void stats_delete(int parc, char *parv[])
{
	USERPTR		user;
	SERVPTR		serv;

	if (parc != 2) {
		writeln(":%s NOTICE %s :Invalid parameter specified", conf.servname, parv[0]);
		writeln(":%s NOTICE %s :Usage: delete (server)", conf.servname, parv[0]);
		return;
	}
	if ((user = user_find(parv[0])) == NULL)
		return;
	if (!user->oper) {
		writeln(":%s NOTICE %s :You are not an IRC Operator", conf.servname, parv[0]);
		return;
	}
	if ((serv = serv_find(parv[1])) == NULL) {
		writeln(":%s NOTICE %s :Server not found", conf.servname, parv[0]);
		return;
	}
	if (serv->connected) {
		writeln(":%s NOTICE %s :Server is currently connected, cannot delete.", conf.servname, parv[0]);
		return;
	}

	serv_rm(serv);
	writeln(":%s NOTICE %s :Deleted server %s", conf.servname, parv[0], parv[1]);
}

void stats_net(int parc, char *parv[])
{
	char	*str;
	int	len;

	if (totals.checks < 1) {
		writeln(":%s NOTICE %s :I have not performed any checks yet.", conf.servname, parv[0]);
		return;
	}

	str = ctime(&start);
	len = strlen(str);
	str[--len] = '\0';

	writeln(":%s NOTICE %s :%s Network Statistics", conf.servname, parv[0], conf.network);
	writeln(":%s NOTICE %s :I Have Performed %d Check(s) Since %s.", conf.servname, parv[0], totals.checks, str);
	writeln(":%s NOTICE %s :Current Users:   %-6d  Max Users:   %-6d  Avg Users: %0.2f", conf.servname, parv[0],
		totals.cur_users, totals.max_users, (double)totals.tot_users/totals.checks);
	writeln(":%s NOTICE %s :Current Dupes:   %-6d  Max Dupes:   %-6d  Avg Dupes: %0.2f", conf.servname, parv[0],
		totals.cur_dupes, totals.max_dupes, (double)totals.tot_dupes/totals.checks);
	writeln(":%s NOTICE %s :Current Opers:   %-6d  Max Opers:   %-6d  Avg Opers: %0.2f", conf.servname, parv[0],
		totals.cur_opers, totals.max_opers, (double)totals.tot_opers/totals.checks);
	writeln(":%s NOTICE %s :Current Servers: %-6d  Max Servers: %-6d  Splits:    %d", conf.servname, parv[0],
		totals.cur_servers, totals.max_servers, totals.splits);
	writeln(":%s NOTICE %s :Server Kills:    %-6d  Oper Kills:  %d", conf.servname, parv[0], totals.skills, totals.okills);
	writeln(":%s NOTICE %s :Domains In RAM:  %d", conf.servname, parv[0], totals.domains);
	writeln(":%s NOTICE %s :Average Number of Users Per Server:   %0.2f", conf.servname, parv[0], totals.cur_servers>1 ?
		(double)totals.cur_users/(totals.cur_servers-1) : (double)0);
}

void stats_stats(int parc, char *parv[])
{
	char		*s;
	SERVPTR		server;
	DOMAINPTR	domain;
	char		*str;
	int		len;

	if (parc != 2) {
		writeln(":%s NOTICE %s :Invalid parameter specified", conf.servname, parv[0]);
		writeln(":%s NOTICE %s :Usage: stats (domain|server)", conf.servname, parv[0]);
		return;
	}

	if (!(s = strchr(parv[1], '.'))) {
		writeln(":%s NOTICE %s :Invalid server/domain specified", conf.servname, parv[0]);
		return;
	}

	if (totals.checks < 1) {
		writeln(":%s NOTICE %s :I have not performed any checks yet", conf.servname, parv[0]);
		return;
	}

	str = ctime(&start);
	len = strlen(str);
	str[--len] = '\0';
	++s;
	if (strchr(s, '.')) { /* Server */
		if ((server = serv_find(parv[1])) == NULL) {
			writeln(":%s NOTICE %s :Server not found", conf.servname, parv[0]);
			return;
		}

		writeln(":%s NOTICE %s :%s Stats For Server: %s", conf.servname, parv[0], conf.network, parv[1]);
		writeln(":%s NOTICE %s :I Have Performed %d Check(s) Since %s.", conf.servname, parv[0], totals.checks, str);
		writeln(":%s NOTICE %s :Current Users:   %-6d (%05.2f%%)  Max Users:  %-6d (%05.2f%%)  Avg Users: %-9.2f (%05.2f%%)",
			conf.servname, parv[0], server->cur_users, totals.cur_users ?
			(double)server->cur_users/totals.cur_users*100.0 : (double)0, server->max_users, totals.max_users ?
			(double)server->max_users/totals.max_users*100.0 : (double)0, (double)server->tot_users/totals.checks,
			totals.tot_users ? ((double)server->tot_users/totals.checks)/
			((double)totals.tot_users/totals.checks)*100.0 : (double)0);
		writeln(":%s NOTICE %s :Current Dupes:   %-6d (%05.2f%%)  Max Dupes:  %-6d (%05.2f%%)  Avg Dupes: %-9.2f (%05.2f%%)",
			conf.servname, parv[0], server->cur_dupes, totals.cur_dupes ?
			(double)server->cur_dupes/totals.cur_dupes*100.0 : (double)0, server->max_dupes, totals.max_dupes ?
			(double)server->max_dupes/totals.max_dupes*100.0 : (double)0, (double)server->tot_dupes/totals.checks,
			totals.tot_dupes ?  ((double)server->tot_dupes/totals.checks)/
			((double)totals.tot_dupes/totals.checks)*100.0 : (double)0);
		writeln(":%s NOTICE %s :Current Opers:   %-6d (%05.2f%%)  Max Opers:  %-6d (%05.2f%%)  Avg Opers: %-9.2f (%05.2f%%)",
			conf.servname, parv[0], server->cur_opers, totals.cur_opers ?
			(double)server->cur_opers/totals.cur_opers*100.0 : (double)0, server->max_opers, totals.max_opers ?
			(double)server->max_opers/totals.max_opers*100.0 : (double)0, (double)server->tot_opers/totals.checks,
			totals.tot_opers ?  ((double)server->tot_opers/totals.checks)/
			((double)totals.tot_opers/totals.checks)*100.0 : (double)0);
		writeln(":%s NOTICE %s :Server Kills:    %-15d  Oper Kills: %d", conf.servname, parv[0], server->skills,
			server->okills);
		writeln(":%s NOTICE %s :Servers Carried: %-15d  Uplink:     %s", conf.servname, parv[0],
			server->links, !server->connected ? "Split" : server->uplink ? server->uplink->servername : "");
		writeln(":%s NOTICE %s :I Have Seen It Split %d Times.", conf.servname, parv[0], server->splits);
	} else {
		if ((domain = domain_find(parv[1])) == NULL) {
			writeln(":%s NOTICE %s :Domain not found", conf.servname, parv[0]);
			return;
		}
		writeln(":%s NOTICE %s :%s Stats For Domain: %s", conf.servname, parv[0], conf.network, parv[1]);
		writeln(":%s NOTICE %s :I Have Performed %d Check(s) Since %s.", conf.servname, parv[0], totals.checks, str);
		writeln(":%s NOTICE %s :Current Users:     %-6d (%05.2f%%)  Max Users: %-6d (%05.2f%%)  Avg Users: %-9.2f (%05.2f%%)",
			conf.servname, parv[0], domain->cur_users, totals.cur_users ? (double)domain->cur_users/totals.cur_users*100.0 :
			(double)0, domain->max_users, totals.max_users ? (double)domain->max_users/totals.max_users*100.0 :
			(double)0, (double)domain->tot_users/totals.checks, totals.tot_users ?
			((double)domain->tot_users/totals.checks)/((double)totals.tot_users/totals.checks)*100.0 : (double)0);
		writeln(":%s NOTICE %s :Current Dupes:     %-6d (%05.2f%%)  Max Dupes: %-6d (%05.2f%%)  Avg Dupes: %-9.2f (%05.2f%%)",
			conf.servname, parv[0], domain->cur_dupes, totals.cur_dupes ? (double)domain->cur_dupes/totals.cur_dupes*100.0 :
			(double)0, domain->max_dupes, totals.max_dupes ? (double)domain->max_dupes/totals.max_dupes*100.0 :
			(double)0, (double)domain->tot_dupes/totals.checks, totals.tot_dupes ?
			((double)domain->tot_dupes/totals.checks)/((double)totals.tot_dupes/totals.checks)*100.0 : (double)0);
		writeln(":%s NOTICE %s :Current Opers:     %-6d (%05.2f%%)  Max Opers: %-6d (%05.2f%%)  Avg Opers: %-9.2f (%05.2f%%)",
			conf.servname, parv[0], domain->cur_opers, totals.cur_opers ? (double)domain->cur_opers/totals.cur_opers*100.0 :
			(double)0, domain->max_opers, totals.max_opers ? (double)domain->max_opers/totals.max_opers*100.0 :
			(double)0, (double)domain->tot_opers/totals.checks, totals.tot_opers ?
			((double)domain->tot_opers/totals.checks)/((double)totals.tot_opers/totals.checks)*100.0 : (double)0);
		writeln(":%s NOTICE %s :Kills From Domain: %d", conf.servname, parv[0], domain->kills);
	}
}

void stats_domain(int parc, char *parv[])
{
	DOMAINPTR	domain;
	LISTPTR		list;
	USERPTR		user;

	if (parc != 2) {
		writeln(":%s NOTICE %s :Invalid parameter specified", conf.servname, parv[0]);
		writeln(":%s NOTICE %s :Usage: domain (domain)", conf.servname, parv[0]);
		return;
	}

	if ((domain = domain_find(parv[1])) == NULL) {
		writeln(":%s NOTICE %s :Domain not found", conf.servname, parv[0]);
		return;
	}

	for (list = domain->users; list != NULL; list = list->next) {
		user = list->data;
		if (user != NULL)
			writeln(":%s NOTICE %s :%s [%s@%s on server %s]", conf.servname, parv[0],
				user->nickname, user->username, user->hostname, user->server->servername);
	}
}

void stats_help(int parc, char *parv[])
{
	USERPTR		user;
	int		oper=0;

	if (((user = user_find(parv[0])) != NULL)&&(user->oper))
		oper=1;

	writeln(":%s NOTICE %s :Welcome, %s, I Support the Following Commands:", conf.servname, parv[0], parv[0]);
	writeln(":%s NOTICE %s :net           - Show Current Network Statistics.", conf.servname, parv[0]);
	writeln(":%s NOTICE %s :missing       - Show Currently Split Servers.", conf.servname, parv[0]);
	writeln(":%s NOTICE %s :map           - Show Map of Network.", conf.servname, parv[0]);
	writeln(":%s NOTICE %s :stats <param> - Show Stats for a Server or Domain.", conf.servname, parv[0]);
	if (oper)
		writeln(":%s NOTICE %s :delete <param> - Delete Stats for a Server.", conf.servname, parv[0]);
	writeln(":%s NOTICE %s :help          - This Text, Duh!.", conf.servname, parv[0]);
} 

void stats_missing(int parc, char *parv[])
{
	int		i, missing=0;
	LISTPTR		cur_ptr;
	SERVPTR		serv;

	writeln(":%s NOTICE %s :Tracking missing servers...", conf.servname, parv[0]);
	for (i=0; i<conf.serverhsize; ++i) {
		for (cur_ptr=servers[i]; cur_ptr!=NULL; cur_ptr=cur_ptr->next) {
			if ((serv = cur_ptr->data) == NULL)
				continue;
			if (!serv->connected) {
				writeln(":%s NOTICE %s :%s", conf.servname, parv[0], serv->servername);
				++missing;
			}
		}
	}

	if (missing)
		writeln(":%s NOTICE %s :----------", conf.servname, parv[0]);
	else
		writeln(":%s NOTICE %s :All servers present and accounted for.", conf.servname, parv[0]);
}

void stats_map(int parc, char *parv[])
{
	int		i;
	LISTPTR		cur_ptr;
	SERVPTR		serv=NULL;

	writeln(":%s NOTICE %s :Generating server map...", conf.servname, parv[0]);
	for (i=0; i<conf.serverhsize; ++i) {
		for (cur_ptr=servers[i]; cur_ptr!=NULL; cur_ptr=cur_ptr->next) {
			serv=cur_ptr->data;
			if ((serv!=NULL)&&(serv->uplink==NULL)&&(serv->connected)) {
				i=conf.serverhsize;
				break;
			}
		}
	}
	dump_map(parv[0], serv, buf);
}

/* parv[0] = sender
 * parv[1] = destination (ME!)
 * parv[2] = message
 */
/* I'm going to be lazy and not bother with the ircd-style command lookup
 * for now.  We don't have enough commands to justify the work. */
void stats_msg(int parc, char *parv[])
{
	int got=0;
	int i=1;
	char *args[3], *cmd, *s;

	args[0] = parv[0];
	if (parc < 3) {
		writeln(":%s NOTICE %s :No command specified", conf.servname, parv[0]);
		return;
	}

	cmd = parv[2];
	s = strchr(cmd, ' ');
	if (s != NULL) {
		*s = '\0';
		++s;
		args[i++] = s;
	}
	args[i] = NULL;

	switch (*cmd) {
		case 'd':
			if (!strcmp(parv[2], "domain")) {
				got=1;
				stats_domain(i, args);
			} else if (!strcmp(parv[2], "delete")) {
				got=1;
				stats_delete(i, args);
			}
			break;
		case 'h':
			if (!strcmp(parv[2], "help")) {
				got=1;
				stats_help(i, args);
			}
			break;
		case 'm':
			if (!strcmp(parv[2], "missing")) {
				got=1;
				stats_missing(i, args);
			} else if (!strcmp(parv[2], "map")) {
				got=1;
				stats_map(i, args);
			}
			break;
		case 'n':
			if (!strcmp(parv[2], "net")) {
				got=1;
				stats_net(i, args);
			}
			break;
		case 's':
			if (!strcmp(parv[2], "stats")) {
				got=1;
				stats_stats(i, args);
			}
			break;
	}
	if (!got)
		writeln(":%s NOTICE %s :Bad command specified", conf.servname, parv[0]);
}

void check_old(void)
{
	LISTPTR		cur_ptr, next_ptr;
	DOMAINPTR	domain;
	SERVPTR		server;
	int		i;

	for (i=0; i<conf.serverhsize; ++i) {
		for (cur_ptr=servers[i]; cur_ptr!=NULL; cur_ptr=next_ptr) {
			next_ptr = cur_ptr->next;
			if ((server=cur_ptr->data) == NULL)
				continue;
			if ((!server->connected)&&((now - server->last) > conf.serverexpire))
				serv_rm(server);
		}
	}

	for (i=0; i<conf.domainhsize; ++i) {
		for (cur_ptr=domains[i]; cur_ptr!=NULL; cur_ptr=next_ptr) {
			next_ptr = cur_ptr->next;
			if ((domain=cur_ptr->data) == NULL)
				continue;
			if ((!domain->cur_users)&&((now - domain->last) > conf.domainexpire))
				domain_del(domain->domainname);
		}
	}
}

void mywrite(int fd, void *ptr, size_t size)
{
	if (iovs == (IOV_MAX-1)) {
		writev(fd, iov, IOV_MAX);
		iovs = 0;
	}
	iov[iovs].iov_base = ptr;
	iov[iovs++].iov_len = size;
}

void write_data(void)
{
	LISTPTR		cur_ptr;
	SERVPTR		server;
	DOMAINPTR	domain;
	int		fd;
	unsigned int	datver=DAT_VER;
	char		tmpnam[PATH_MAX+1];
	char		tmpnam2[PATH_MAX+1];
	int		i;

	snprintf(tmpnam, PATH_MAX+1, "%s.new", conf.datafile);
	snprintf(tmpnam2, PATH_MAX+1, "%s.old", conf.datafile);

	if ((fd = open(tmpnam, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) < 0)
		die("Error opening file: %s\n", tmpnam);

	totals.tot_users += totals.cur_users;
	totals.tot_opers += totals.cur_opers;
	totals.tot_dupes += totals.cur_dupes;
	++totals.checks;

	mywrite(fd, "AlcStats", 8);
	mywrite(fd, &datver, sizeof(unsigned long)); /* And our conf version */
	mywrite(fd, &totals, sizeof(struct total_struct));
#ifdef USE_RRD
	updaterrd(conf.network, totals.cur_users, totals.cur_dupes, totals.cur_opers);
#endif
	for (i=0; i<conf.serverhsize; ++i) {
		for (cur_ptr=servers[i]; cur_ptr!=NULL; cur_ptr=cur_ptr->next) {
			if ((server=cur_ptr->data) == NULL)
				continue;
			server->tot_users += server->cur_users;
			server->tot_opers += server->cur_opers;
			server->tot_dupes += server->cur_dupes;
			mywrite(fd, server, sizeof(struct server_struct));
#ifdef USE_RRD
			updaterrd(server->servername, server->cur_users, server->cur_dupes, server->cur_opers);
#endif
		}
	}
	for (i=0; i<conf.domainhsize; ++i) {
		for (cur_ptr=domains[i]; cur_ptr!=NULL; cur_ptr=cur_ptr->next) {
			if ((domain=cur_ptr->data) == NULL)
				continue;
			domain->tot_users += domain->cur_users;
			domain->tot_opers += domain->cur_opers;
			domain->tot_dupes += domain->cur_dupes;
			mywrite(fd, domain, sizeof(struct domain_struct));
		}
	}
	if (iovs) {
		writev(fd, iov, iovs);
		iovs = 0;
	}
	close(fd);
	unlink(tmpnam2);
	rename(conf.datafile, tmpnam2);
	rename(tmpnam, conf.datafile);
}

void read_data(void)
{
	SERVPTR			serv_ptr;
	DOMAINPTR		dom_ptr;
	FILE			*fp;
	char			servname[HOSTLEN+1];
	char			domname[HOSTLEN+1];
	char			header[8+1];
	unsigned int		size=0, count=0, datver=0;
	unsigned short int	doms, servs;
	int			offset;
	SERVERS			serv;
	DOMAINS			dom;

	if ((fp = fopen(conf.datafile, "rb")) == NULL) {
		fprintf(stderr, "Error opening stats file: %s (continuing with fresh stats...)\n", conf.datafile);
		return;
	}

	fread(header, 8*sizeof(char), 1, fp); /* Try to find our header */
	header[8] = '\0';
	if (memcmp(header, "AlcStats", 8)) {
		fprintf(stderr, ".dat files older than 3.2 are no longer supported. (continuing with fresh stats...)\n");
		return;
	}
	fread(&datver, sizeof(unsigned int), 1, fp);
	if (datver < 2) {
		fprintf(stderr, ".dat files older than 3.2 are no longer supported. (continuing with fresh stats...)\n");
		return;
	}
	if (datver >= 3) {
		fread(&totals, sizeof(struct total_struct), 1, fp);
		servs = totals.tot_servers;
		doms = totals.domains;
		totals.cur_users = totals.cur_opers = totals.cur_dupes = totals.tot_servers = totals.domains = 0;
		totals.cur_servers = 1;
	} else {
		fread(&(totals.checks), sizeof(unsigned int), 1, fp);
		fread(&(totals.tot_users), sizeof(unsigned long int), 1, fp);
		fread(&(totals.tot_opers), sizeof(unsigned long int), 1, fp);
		fread(&(totals.tot_dupes), sizeof(unsigned long int), 1, fp);
		fread(&(totals.max_users), sizeof(unsigned int), 1, fp);
		fread(&(totals.max_opers), sizeof(unsigned int), 1, fp);
		fread(&(totals.max_dupes), sizeof(unsigned int), 1, fp);
		fread(&(totals.splits), sizeof(unsigned short int), 1, fp);
		fread(&(totals.skills), sizeof(unsigned int), 1, fp);
		fread(&(totals.okills), sizeof(unsigned int), 1, fp);
		fread(&(totals.max_servers), sizeof(unsigned short int), 1, fp);
		fread(&(servs), sizeof(unsigned short int), 1, fp);
		fread(&(doms), sizeof(unsigned short int), 1, fp);
		/* We don't read-in "current" data, so just skip over it */
		offset = ((sizeof(unsigned int)*3)+sizeof(unsigned short int));
		fseek(fp, offset, SEEK_CUR);
	}
	for (count=0; count < servs; count++) {
		if (datver >= 3) {
			fread(&serv, sizeof(struct server_struct), 1, fp);
			serv_ptr = serv_add(serv.servername);
			serv_ptr->tot_users = serv.tot_users;
			serv_ptr->tot_opers = serv.tot_opers;
			serv_ptr->tot_dupes = serv.tot_dupes;
			serv_ptr->max_users = serv.max_users;
			serv_ptr->max_opers = serv.max_opers;
			serv_ptr->max_dupes = serv.max_dupes;
			serv_ptr->splits = serv.splits;
			serv_ptr->okills = serv.okills;
			serv_ptr->skills = serv.skills;
			serv_ptr->last = serv.last;
		} else {
			fread(&size, sizeof(unsigned int), 1, fp);
			fread(servname, size, 1, fp);
			serv_ptr = serv_add(servname);
			fread(&(serv_ptr->tot_users), sizeof(unsigned long int), 1, fp);
			fread(&(serv_ptr->tot_opers), sizeof(unsigned long int), 1, fp);
			fread(&(serv_ptr->tot_dupes), sizeof(unsigned long int), 1, fp);
			fread(&(serv_ptr->max_users), sizeof(unsigned short), 1, fp);
			fread(&(serv_ptr->max_opers), sizeof(unsigned short), 1, fp);
			fread(&(serv_ptr->max_dupes), sizeof(unsigned short), 1, fp);
			fread(&(serv_ptr->splits), sizeof(unsigned short), 1, fp);
			fread(&(serv_ptr->okills), sizeof(unsigned int), 1, fp);
			fread(&(serv_ptr->skills), sizeof(unsigned int), 1, fp);
			fread(&(serv_ptr->last), sizeof(time_t), 1, fp);
			/* We don't read-in "current" data, so just skip over it */
			offset = ((sizeof(unsigned short int)*3));
			fseek(fp, offset, SEEK_CUR);
			fread(&size, sizeof(unsigned int), 1, fp);
			if (size)
				fseek(fp, size, SEEK_CUR);
		}
		serv_ptr->connected = 0;
	}
	for (count = 0; count < doms; count++) {
		if (datver >= 3) {
			fread(&dom, sizeof(struct domain_struct), 1, fp);
			dom_ptr = domain_add(dom.domainname);
			dom_ptr->tot_users = dom.tot_users;
			dom_ptr->tot_dupes = dom.tot_dupes;
			dom_ptr->tot_opers = dom.tot_opers;
			dom_ptr->max_users = dom.max_users;
			dom_ptr->max_dupes = dom.max_dupes;
			dom_ptr->max_opers = dom.max_opers;
			dom_ptr->kills = dom.kills;
			dom_ptr->last = dom.last;
		} else {
			fread(&size, sizeof(unsigned int), 1, fp);
			fread(domname, size, 1, fp);
			dom_ptr = domain_add(domname);
			fread(&(dom_ptr->tot_users), sizeof(unsigned long int), 1, fp);
			fread(&(dom_ptr->tot_dupes), sizeof(unsigned long int), 1, fp);
			fread(&(dom_ptr->tot_opers), sizeof(unsigned long int), 1, fp);
			fread(&(dom_ptr->max_users), sizeof(unsigned short), 1, fp);
			fread(&(dom_ptr->max_dupes), sizeof(unsigned short), 1, fp);
			fread(&(dom_ptr->max_opers), sizeof(unsigned short), 1, fp);
			fread(&(dom_ptr->kills), sizeof(unsigned int), 1, fp);
			fread(&(dom_ptr->last), sizeof(time_t), 1, fp);
			offset = ((sizeof(unsigned short int)*3));
			fseek(fp, offset, SEEK_CUR);
		}
	}
	fclose(fp);
}

void init_totals(void)
{
	memset(&totals, 0, sizeof(struct total_struct));
	totals.cur_servers=1;
	totals.max_servers=1;
}

