/* conf file parser, 'borrowed' from ircd protocol parser. */

#include "alcstats.h"
#include "parse.h"
#include "list.h"
#include "connlist.h"
#include "exempt.h"
#define CONFTAB
#include "conf.h"

struct confitems conf;

void init_conf_parse(struct Message *mptr)
{
	int		i;
	struct Message	*mpt = mptr;

	for (i=0; mpt->cmd; mpt++)
		i++;
	qsort((void *)mptr, i, sizeof(struct Message),
	      (int (*)(const void *, const void *)) mcmp);
	conf_tree_root = (MESSAGE_TREE *)MyMalloc(sizeof(MESSAGE_TREE));
	mpt = do_msg_tree(conf_tree_root, "", mptr);
}

static struct Message *conf_parse(char *cmd)
{
	char		r;
	MESSAGE_TREE	*mtree = conf_tree_root;

	while ((r = *cmd++)) { 
		r &= 0xdf;
		if (r < 'A' || r > 'Z')
			return NULL;
		mtree = mtree->pointers[r - 'A'];
		if (!mtree)
			return NULL;
		if (mtree->final && !strcasecmp(mtree->final, cmd))
			return mtree->msg;
	}
	return ((struct Message *)NULL);
}


static inline int string_to_array(char *string, char *parv[MAXPARA])
{
	char *p;
	char *buf = string;
	int x = 0;

	parv[x] = NULL;

	do {
		parv[x++] = buf;
		parv[x]   = NULL;

		if ((p = strchr(buf, ':')) != NULL) {
			*p++ = '\0';
			buf  = p;
		} else
			return(x);

		if (*buf == '\0')
			return(x);
	} while (x < MAXPARA - 1);

	parv[x++] = p;
	parv[x]   = NULL;
	return(x);
}

/* Modified parse from ircd - used to parse config file */
void confparse(char *line)
{
	char		*s, *para[MAXPARA+1];
	unsigned int	i = 0;
	struct Message	*mptr;

	if (*line == '\0')
		return;

	if (*line == '#')
		return;

	if ((s = strchr(line, ':')) != NULL)
		*s++ = '\0';

	mptr = conf_parse(line);

	if ((mptr == NULL) || (mptr->cmd == NULL))
		die("Unknown %s line in conf file\n", line);

	if (s != NULL)
		i = string_to_array(s, para);
	else {  
		i = 0;
		para[0] = NULL;
	}

	if (i != mptr->parameters)
		die("Bad %s line in conf file\n", mptr->cmd);

	(*mptr->func)(i, para);
}

void do_conf(char *conf)
{
	char	line[1024+1];
	FILE	*cp;
	int	len;

	if ((cp = fopen(conf, "r")) == NULL) {
		fprintf(stderr, "Unable to open config file!\n");
		exit(2);
	}

	while (fgets(line, 1024+1, cp) != NULL) {
		len = strlen(line);
		if (line[len-1] == '\n')
			line[len-1] = '\0';
		confparse(line);
	}
}

void c_admin(int argc, char **argv)
{
	conf.admin = mystrdup(argv[0]);
}

void c_calc(int argc, char **argv)
{
	conf.calctime = atoi(argv[0]);
}

void c_connect(int argc, char **argv)
{
	int port;
	port = atoi(argv[3]);
	conn_add(argv[0], argv[1], argv[2], port);
}

void c_datafile(int argc, char **argv)
{
	conf.datafile = mystrdup(argv[0]);
}

void c_description(int argc, char **argv)
{
	conf.servdesc = mystrdup(argv[0]);
}

void c_domainexpire(int argc, char **argv)
{
	conf.domainexpire = atoi(argv[0]);
}

void c_domainhash(int argc, char **argv)
{
	conf.domainhsize = atoi(argv[0]);
}

void c_domainheap(int argc, char **argv)
{
	conf.domainheap = atoi(argv[0]);
}

void c_exempt(int argc, char **argv)
{
	exempt_add(argv[0]);
}

void c_location(int argc, char **argv)
{
	conf.location = mystrdup(argv[0]);
}

void c_maxthresh(int argc, char **argv)
{
	conf.maxthresh = atoi(argv[0]);
}

void c_network(int argc, char **argv)
{
	conf.network = mystrdup(argv[0]);
}

void c_rrddir(int argc, char **argv)
{
	conf.rrddir = mystrdup(argv[0]);
}

void c_serverexpire(int argc, char **argv)
{
	conf.serverexpire = atoi(argv[0]);
}

void c_serverhash(int argc, char **argv)
{
	conf.serverhsize = atoi(argv[0]);
}

void c_serverheap(int argc, char **argv)
{
	conf.serverheap = atoi(argv[0]);
}

void c_servname(int argc, char **argv)
{
	conf.servname = mystrdup(argv[0]);
}

void c_userhash(int argc, char **argv)
{
	conf.userhsize = atoi(argv[0]);
}

void c_userheap(int argc, char **argv)
{
	conf.userheap = atoi(argv[0]);
}

void c_username(int argc, char **argv)
{
	conf.username = mystrdup(argv[0]);
}

