#include "alcstats.h"
#include "list.h"
#include "hash.h"
#include "balloc.h"
#include "servlist.h"
#include "domainlist.h"
#ifdef USE_RRD
#include "rrdtool.h"
#endif

#ifndef NOBALLOC
BlockHeap	*domain_heap;
#endif
LISTPTR		*domains;

DOMAINPTR domain_add(char *domain)
{
	LISTPTR		ptr;
	DOMAINPTR	dom;
	unsigned int	key;

	key = hash_domain(domain);
#ifndef NOBALLOC
	dom = BlockHeapAlloc(domain_heap);
#else
	dom = MyMalloc(sizeof(DOMAINS));
#endif
	ptr = list_add_ptr(&(domains[key]), dom);
	mystrlcpy(dom->domainname, domain, HOSTLEN);
	dom->last = now;
	++totals.domains;
	return dom;
}

void domain_del(char *domain)
{
	DOMAINPTR	dom=NULL;
	LISTPTR		list, next_ptr;
	unsigned int	key;

#ifdef USE_RRD
	deleterrd(domain);
#endif
	key = hash_domain(domain);
	if ((dom = domain_find(domain)) == NULL)
		return;
	list = dom->users;
	while (list != NULL) {
		next_ptr = list->next;
#ifndef NOBALLOC
		BlockHeapFree(list_heap, list);
#else
		MyFree(list);
#endif
		if (next_ptr == NULL)
			break;
		list = next_ptr;
	}
	list_del_ptr(&(domains[key]), dom);
#ifndef NOBALLOC
	BlockHeapFree(domain_heap, dom);
#else
	MyFree(dom);
#endif
	--totals.domains;
}

DOMAINPTR domain_find(char *domain)
{
	LISTPTR		ptr;
	DOMAINPTR	dom;
	unsigned int	key;

	key = hash_domain(domain);
	for (ptr = domains[key]; ptr != NULL; ptr = ptr->next) {
		if ((dom = ptr->data) == NULL)
			continue;
		if (!strcmp(dom->domainname, domain))
			return(dom);
	}

	return NULL;
}

void domain_init(void)
{
	int	i;

#ifndef NOBALLOC
	/* If we didn't specify a size, try to guess one */
	if (!conf.domainheap)
		conf.domainheap = (int)((double)conf.domainhsize*0.95);
	domain_heap = BlockHeapCreate(sizeof(DOMAINS), conf.domainheap);
#endif
	domains = MyMalloc(sizeof(LISTPTR)*conf.domainhsize);
	for (i=0; i<conf.domainhsize; ++i)
		domains[i] = NULL;
}

