#include "alcstats.h"
#include "list.h"
#include "hash.h"
#include "balloc.h"
#include "servlist.h"
#include "userlist.h"

#ifndef NOBALLOC
BlockHeap	*list_heap;
#endif

LISTPTR list_add_ptr(LISTPTR *list, void *data)
{
        LISTPTR	tmp_ptr = *list;

#ifndef NOBALLOC
	*list = BlockHeapAlloc(list_heap);
#else
	*list = MyMalloc(sizeof(List));
#endif
	(*list)->data = data;
        (*list)->next = tmp_ptr;
        return *list;
}

void list_add_link(void *data, LISTPTR node, LISTPTR *list)
{
	LISTPTR tmp_ptr = *list;

	*list = node;
	(*list)->data = data;
	(*list)->next = tmp_ptr;
	return;
}

void list_del_ptr(LISTPTR *list, void *data)
{
        LISTPTR	cur_ptr, prev_ptr=NULL;

        for (cur_ptr = *list; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
		if (cur_ptr->data == data)
                        break;
                prev_ptr = cur_ptr;
        }

        if (cur_ptr == NULL)
                return;

	if (prev_ptr == NULL)
		*list = cur_ptr->next;
	else
		prev_ptr->next = cur_ptr->next;

#ifndef NOBALLOC
	BlockHeapFree(list_heap, cur_ptr);
#else
	MyFree(cur_ptr);
#endif
}

void list_del_link(LISTPTR ptr, LISTPTR *list)
{
        LISTPTR	cur_ptr, prev_ptr=NULL;

        for (cur_ptr = *list; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
		if (cur_ptr == ptr)
                        break;
                prev_ptr = cur_ptr;
        }

        if (cur_ptr == NULL)
                return;

	if (prev_ptr == NULL)
		*list = cur_ptr->next;
	else
		prev_ptr->next = cur_ptr->next;
}

/* Not used right now - Maybe in the future? */
LISTPTR list_find_ptr(LISTPTR *list, void *data)
{
	LISTPTR	cur_ptr;

	for (cur_ptr = *list; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
		if (cur_ptr->data == data)
			return(cur_ptr);
	}
	return NULL;
}

int dupe_check(LISTPTR list, const char *username, const char *hostname)
{
	LISTPTR	cur_ptr;
	USERPTR	user;
	int	count=0;

	for (cur_ptr = list; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
		if (cur_ptr->data == NULL)
			continue;
		user = cur_ptr->data;
		if ((!strcmp(user->username, username))&&(!strcmp(user->hostname, hostname)))
			++count;
	}
	return count;
}

void unmark_dupe(LISTPTR list, const char *username, const char *hostname)
{
	LISTPTR	cur_ptr;
	USERPTR	user;

	for (cur_ptr = list; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
		if (cur_ptr->data == NULL)
			continue;
		user = cur_ptr->data;
		if ((user->dupe)&&(!strcmp(user->username, username))&&(!strcmp(user->hostname, hostname))) {
			user->dupe = 0;
			if (user->server->cur_dupes)
				--user->server->cur_dupes;
			return;
		}
	}
}

void list_init(void)
{
#ifndef NOBALLOC
	list_heap = BlockHeapCreate(sizeof(List), LIST_HEAP_SIZE);
#endif
}

