/* parser and command lookup code 'borrowed' from ircd.
 * client<->server protcol has been removed. */

#include "alcstats.h"
#include "list.h"
#include "connlist.h"
#define MSGTAB
#include "parse.h"

struct Message *do_msg_tree(MESSAGE_TREE *mtree, char *prefix,
				   struct Message *mptr)
{
	char		newpref[64];  /* must be longer than any command */
	int		c, c2, lp;
	MESSAGE_TREE	*mtree1;

	lp = strlen(prefix);
	if (!lp || !strncmp(mptr->cmd, prefix, lp)) { 
		if (!mptr[1].cmd || (lp && strncmp(mptr[1].cmd, prefix, lp))) { 
			/* non ambiguous -> make a final case */
			mtree->final = mptr->cmd + lp;
			mtree->msg = mptr;
			for (c=0; c<=25; c++)
				mtree->pointers[c] = NULL;
			return mptr+1;
		} else {
			/* ambigous -> make new entries for each of the letters that match */
			if (!strcasecmp(mptr->cmd, prefix)) {
				/* fucking OPERWALL blows me */
				mtree->msg = mptr;
				mptr++;
			} else
				mtree->final = NULL;

			for (c='A'; c<='Z'; c++) {
				if (mptr->cmd[lp] == c) {
					mtree1 = (MESSAGE_TREE *)MyMalloc(sizeof(MESSAGE_TREE));
					mtree1->final = NULL;
					mtree->pointers[c-'A'] = mtree1;
					strcpy(newpref, prefix);
					newpref[lp] = c;
					newpref[lp+1] = '\0';
					mptr = do_msg_tree(mtree1, newpref, mptr);
					if (!mptr->cmd || strncmp(mptr->cmd, prefix, lp)) {
						for (c2=c+1-'A'; c2<=25; c2++)
							mtree->pointers[c2] = NULL;
						return mptr;
					}
				} else {
					mtree->pointers[c-'A'] = NULL;
				}
			}
			return mptr;
		}
	}
	return (0);
}

int mcmp(struct Message *m1, struct Message *m2)
{ 
	return strcmp(m1->cmd, m2->cmd);
}

void init_tree_parse(struct Message *mptr)
{
	int		i;
	struct Message	*mpt = mptr;

	for (i=0; mpt->cmd; mpt++)
		i++;
	qsort((void *)mptr, i, sizeof(struct Message),
	      (int (*)(const void *, const void *)) mcmp);
	msg_tree_root = (MESSAGE_TREE *)MyMalloc(sizeof(MESSAGE_TREE));
	mpt = do_msg_tree(msg_tree_root, "", mptr);
}

static struct Message *tree_parse(char *cmd)
{
	char		r;
	MESSAGE_TREE	*mtree = msg_tree_root;

	while ((r = *cmd++)) { 
		r &= 0xdf;
		if (r < 'A' || r > 'Z')
			return NULL;
		mtree = mtree->pointers[r - 'A'];
		if (!mtree)
			return NULL;
/*		if (mtree->final == (void *)1) { 
			if (!*cmd)
				return mtree->msg;
		} else*/ if (mtree->final && !strcasecmp(mtree->final, cmd))
			return mtree->msg;
	}
	return ((struct Message *)NULL);
}


static inline int string_to_array(char *string, char *parv[MAXPARA])
{
	char *p;
	char *buf = string;
	int x = 1;

	parv[x] = NULL;

	while (*buf == ' ') /* skip leading spaces */
		buf++;

	if (*buf == '\0') /* ignore all-space args */
		return(x);

	do {
		if (*buf == ':') { /* Last parameter */
			buf++;
			parv[x++] = buf;
			parv[x]   = NULL;
			return(x);
		} else {
			parv[x++] = buf;
			parv[x]   = NULL;

			if ((p = strchr(buf, ' ')) != NULL) {
				*p++ = '\0';
				buf  = p;
			} else
				return(x);
		}

		while (*buf == ' ')
			buf++;

		if (*buf == '\0')
			return(x);
	} while (x < MAXPARA - 1);

	if (*p == ':')
		p++;

	parv[x++] = p;
	parv[x]   = NULL;
	return(x);
}

/* Modified parse from ircd - client<->server protocol parsing removed and
 * extra paranoid added by me :) */
void parse(char *line)
{
	char		*ch, *s, *sender, *para[MAXPARA+1];
	unsigned int	i = 0, numeric;
	struct Message	*mptr;
	CONNPTR		conn;

	if (opts.debug) {
		printf("%s\n", line);
		fflush(stdout);
	}
	if (opts.logfile != NULL) {
		write(logfd, line, strlen(line));
		write(logfd, "\n", 1);
	}

	if (*line == '\0')
		return;

	for (ch = line; *ch == ' '; ch++) /* skip spaces */
		;

	if (*ch == ':') {
		ch++;
		sender = ch;

		if ((s = strchr(ch, ' ')) != NULL) {
			*s = '\0';
			s++;
			ch = s;
		}

		para[0] = sender;

		while (*ch == ' ')
			ch++;
	} else {
		conn = cur_conn->data;
		if ((conn != NULL)&&(conn->name != NULL))
			para[0] = conn->name;
		else
			para[0] = "";
	}

	if (*ch == '\0')
		return;

	if ((s = strchr(ch, ' ')) != NULL)
		*s++ = '\0';

	/* Yes this is hackish, but we only parse one numeric right now */
	if (isdigit(ch[0])&&isdigit(ch[1])&&isdigit(ch[2])) {
		numeric = atoi(ch);
		if (s != NULL)
			i = string_to_array(s, para);
		else {  
			i = 0;
			para[1] = NULL;
		}
		switch (numeric) {
			case 351:
				m_351(i, para);
				break;
			default:
				break;
		}
		return;
	}

	mptr = tree_parse(ch);

	if ((mptr == NULL) || (mptr->cmd == NULL))
		return;

	if (s != NULL)
		i = string_to_array(s, para);
	else {  
		i = 0;
		para[1] = NULL;
	}

	(*mptr->func)(i, para);
}

