#include "alcstats.h"
#include "serv.h"
#include "parse.h"
#include "list.h"
#include "hash.h"
#include "balloc.h"
#include "servlist.h"
#include "userlist.h"
#include "domainlist.h"
#include "connlist.h"
#include "exempt.h"
#ifdef USE_RRD
#include "rrdtool.h"
#endif

char			interface[USERLEN + HOSTLEN + 2];
char			logbuf[LOGSIZE+1];
int			logfd;
struct net_struct	net;
static time_t		lastp;
static unsigned char	got_ping=0;
time_t			now;
char			connected;

/* m_nick
 * parv[0] = NULL when new user; old nick when nick change
 * parv[1] = (new) nickname
 * parv[2] = hopcount when new user; TS when nick change
 * parv[3] = TS
 * parv[4] = umode
 * parv[5] = username
 * parv[6] = hostname
 * parv[7] = server
 * parv[8] = ircname
 */
void m_nick(int parc, char *parv[])
{
	SERVPTR		server;
	USERPTR		user, newuser;
	DOMAINPTR	dom;
	char		domain[HOSTLEN+1];

	if (parc == 9) {
		if ((server = serv_find(parv[7])) == NULL) {
			fprintf(stderr, "Can't find existing server for user?!\n");
			return;
		}

		if ((user = user_add(parv[1], parv[5], parv[6], server)) == NULL)
			return;
		if (++totals.cur_users > totals.max_users) {
			totals.max_users=totals.cur_users;
			if ((totals.checks)&&(!(totals.max_users % conf.maxthresh)))
				writeln(":%s WALLOPS :New Max Network Clients: %d", conf.servname, totals.max_users);
		}
		if (++server->cur_users > server->max_users)
			server->max_users=server->cur_users;

		if (host_to_domain(parv[6], domain) != NULL) {
			if ((dom = domain_find(domain)) == NULL)
				dom = domain_add(domain);
			if (++dom->cur_users > dom->max_users)
				dom->max_users=dom->cur_users;
			if (dupe_check(dom->users, parv[5], parv[6])) {
				user->dupe = 1;
				if (++totals.cur_dupes > totals.max_dupes)
					totals.max_dupes = totals.cur_dupes;
				if (++server->cur_dupes > server->max_dupes)
					server->max_dupes = server->cur_dupes;
				if (++dom->cur_dupes > dom->max_dupes)
					dom->max_dupes = dom->cur_dupes;
			}
			list_add_ptr(&(dom->users), user);
			user->domain = dom;
		}

		if (strchr(parv[4], 'o')) {
			user->oper = 1;
			if (++totals.cur_opers > totals.max_opers)
				totals.max_opers=totals.cur_opers;
			if (++server->cur_opers > server->max_opers)
				server->max_opers=server->cur_opers;
			if (user->domain)
				if (++user->domain->cur_opers > user->domain->max_opers)
					user->domain->max_opers = user->domain->cur_opers;
		}
	} else if (parc == 3) {
		if (!strcasecmp(parv[0], parv[1])) /* Same nick - ignore */
			return;
		if ((user = user_find(parv[0])) == NULL)
			return;
		if ((newuser = user_add(parv[1], user->username, user->hostname, user->server)) == NULL)
			return;
		if ((dom = user->domain) != NULL) {
			list_add_ptr(&(dom->users), newuser);
			newuser->domain = dom;
		}
		newuser->oper = user->oper;
		newuser->dupe = user->dupe;
		user_del(parv[0]);
	}
}

/* m_privmsg
 * parv[0] = sender
 * parv[1] = destination channels/nicks
 * parv[2] = message text
 */
void m_privmsg(int parc, char *parv[])
{
	if (!strcmp(parv[1], interface))
		stats_msg(parc, parv);
}

/* m_quit
 * parv[0] = sender
 * parv[1] = comment
 */
void m_quit(int parc, char *parv[])
{
	user_signoff(parv[0]);
}

/* m_ping
 * parv[0] = sender
 * parv[1] = origin
 * parv[2] = desination - I'm a leaf so always me!
 */
void m_ping(int parc, char *parv[])
{
	writeln(":%s PONG %s :%s", conf.servname, (parc > 2) ? parv[2] : conf.servname, parv[1]);
	if (!got_ping) {
		got_ping = 1;
		nextc = now + conf.calctime;
		check_old();
		write_data();
	}
}

/* m_kill
 * parv[0] = sender
 * parv[1] = kill victim
 * parv[2] = kill path
 */
void m_kill(int parc, char *parv[])
{
	SERVPTR server;
	USERPTR user;

	if (strstr(parv[0], ".")) {
		if (!is_exempt(parv[0])) {
			++totals.skills;
			server = serv_find(parv[0]);
			if (server != NULL)
				++server->skills;
		}
	} else {
		user = user_find(parv[0]);
		if (user != NULL) {
			if (!is_exempt(user->server->servername)) {
				++totals.okills;
				++user->server->okills;
				if (user->domain != NULL)
					++user->domain->kills;
			}
		} else
			++totals.okills;
	}
	user_signoff(parv[1]);
}

/* m_capab
 * parv[0] = sender
 * parv[1] = capabilities
 */
void m_capab(int parc, char *parv[])
{
	char *s, *ptr;

	if (parc < 2)
		return;

	for (ptr = parv[1]; ptr != NULL; ptr = s) {
		if ((s = strchr(ptr, ' ')) != NULL) {
			*s = '\0';
			++s;
		}
		if (!strcmp(ptr, "QS"))
			opts.qs = 1;
	}
}

/* m_server
 * parv[0] = sender
 * parv[1] = servername
 * parv[2] = hopcount
 * parv[3] = serverinfo
 */
void m_server(int parc, char *parv[])
{
	SERVPTR uplink, downlink;

	if ((parv[0]==NULL)||(parv[0][0]=='\0'))
		parv[0] = conf.servname;

	if ((downlink = serv_find(parv[1])) == NULL)
		downlink = serv_add(parv[1]);

	if (++totals.cur_servers > totals.max_servers)
		totals.max_servers = totals.cur_servers;

	downlink->connected = 1;
	downlink->last = now;

	if (strcmp(parv[0], parv[1])) {
		if ((uplink = serv_find(parv[0])) != NULL) {
			uplink->links++;
			list_add_ptr(&(uplink->servers), downlink);
		}
		downlink->uplink = uplink;
	} else
		downlink->uplink = NULL;

#ifdef USE_RRD
	makerrd(parv[1]);
#endif

	writeln(":%s VERSION :%s", conf.servname, parv[1]);
}

/* m_squit
 * parv[0] = sender
 * parv[1] = disconnected server
 * parv[2] = comment
 */
void m_squit(int parc, char *parv[])
{
	if (!strcmp(parv[1], conf.servname)) { /* I was SQUIT'd */
		close(pollfd[0].fd);
		pollfd[0].fd = -1;
		pollfd[0].events = 0;
		connected = 0;
		clear_all();
		rotate_conn();
		return;
	}

	serv_del(parv[1], 1);
}

/* m_version
 * parv[0] = sender
 * parv[1] = destination - I'm a leaf so always me!
 */
void m_version(int parc, char *parv[])
{
	writeln(":%s 351 %s %s. %s :ACeGHiKMprZ TS5owc", conf.servname, parv[0], AS_VER, conf.servname);
}

/* m_mode
 * parv[0] = sender
 * parv[1] = channel/nick
 */
void m_mode(int parc, char *parv[])
{
	USERPTR	user;
	char	*ptr;
	char	sign=' ';

	if ((parv[1][0]=='#')||(parv[1][0]=='&')||(strchr(parv[1], '.')))
		return;
	if ((parc > 2)&&((ptr = strchr(parv[2], 'o')))) {
		if ((user = user_find(parv[0])) == NULL)
			return;
		for (; ptr >= parv[2]; --ptr) {
			if (*ptr == '+') {
				sign='+';
				break;
			} else if (*ptr == '-') {
				sign='-';
				break;
			}
		}
		if ((sign=='+')&&(!user->oper)) {
			user->oper = 1;
			if (++totals.cur_opers > totals.max_opers)
				totals.max_opers = totals.cur_opers;
			if (++user->server->cur_opers > user->server->max_opers)
				user->server->max_opers = user->server->cur_opers;
			if (user->domain != NULL)
				if (++user->domain->cur_opers > user->domain->max_opers)
					user->domain->max_opers = user->domain->cur_opers;
		} else if ((sign=='-')&&(user->oper)) {
			user->oper = 0;
			--totals.cur_opers;
			--user->server->cur_opers;
			if (user->domain != NULL)
				--user->domain->cur_opers;
		}
	}
}

/* m_stats
 * parv[0] = sender
 * parv[1] = stat
 * parv[2] = destination - I'm a leaf so always me!
 */
void m_stats(int parc, char *parv[])
{
	USERPTR		user;
	DOMAINPTR	domain;
	SERVPTR		serv;
	LISTPTR		cur_ptr;
	int		i;

	if ((user = user_find(parv[0])) == NULL)
		return;
	if (!user->oper) {
		writeln(":%s NOTICE %s :You are not an IRC Operator", conf.servname, parv[0]);
		return;
	}

	switch (*parv[1]) {
		case 'D':
			{
				for (i=0; i<conf.userhsize; ++i) {
					for (cur_ptr = users[i]; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
						if ((user = cur_ptr->data) == NULL)
							continue;
						if (user->dupe)
							writeln(":%s NOTICE %s :%s [%s@%s on server %s]", conf.servname, parv[0],
								user->nickname, user->username, user->hostname,
								user->server->servername);
					}
				}
			}
			break;
		case 'd':
			{
				for (i=0; i<conf.domainhsize; ++i) {
					for (cur_ptr = domains[i]; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
						if ((domain = cur_ptr->data) == NULL)
							continue;
						writeln(":%s NOTICE %s :%s [%d]", conf.servname, parv[0], domain->domainname,
							domain->cur_users);
					}
				}
			}
			break;
		case 's':
			{
				for (i=0; i<conf.serverhsize; ++i) {
					for (cur_ptr = servers[i]; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
						if ((serv = cur_ptr->data) == NULL)
							continue;
						writeln(":%s NOTICE %s :Server: %s, Uplink: %s, Version: %s", conf.servname,
							parv[0], serv->servername, !serv->connected ? "Split" : serv->uplink ?
							serv->uplink->servername : "",
							serv->version[0] != '\0' ? serv->version : "");
					}
				}
			}
			break;
		case 'U':
			{
				for (i=0; i<conf.userhsize; ++i) {
					for (cur_ptr = users[i]; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
						if ((user = cur_ptr->data) == NULL)
							continue;
						writeln(":%s NOTICE %s :%s [%s@%s on server %s]", conf.servname, parv[0],
							user->nickname, user->username, user->hostname, user->server->servername);
					}
				}
			}
			break;
	}
		
}

/* m_admin
 * parv[0] = sender
 * parv[1] = destination - I'm a leaf so always me!
 */
void m_admin(int parc, char *parv[])
{
	writeln(":%s 256 %s :Administrative info about %s", conf.servname, parv[0], conf.servname);
	writeln(":%s 257 %s :%s", conf.servname, parv[0], conf.servdesc);
	writeln(":%s 258 %s :%s", conf.servname, parv[0], conf.location);
	writeln(":%s 259 %s :%s", conf.servname, parv[0], conf.admin);
}

/* m_info
 * parv[0] = sender
 * parv[1] = destination - I'm a leaf so always me!
 */
void m_info(int parc, char *parv[])
{
	writeln(":%s 373 %s :Server INFO", conf.servname, parv[0]);
	writeln(":%s 371 %s :%s", conf.servname, parv[0], AS_VER);
	writeln(":%s 371 %s :%s", conf.servname, parv[0], conf.servdesc);
	writeln(":%s 371 %s :%s", conf.servname, parv[0], "-Dustin Marquess <jailbird@alcatraz.fdf.net>");
	writeln(":%s 374 %s :End of /INFO list.", conf.servname, parv[0]);
}

/* m_351
 * parv[0] = sender
 * parv[1] = me
 * parv[2] = version
 * parv[3] = server name
 * parv[4] = switches
 * parv[5] = TS version
 */
void m_351(int parc, char *parv[])
{
	int	len;
	SERVPTR	serv;

	if ((parv[0]==NULL)||(parv[0][0]=='\0'))
		return;
	if ((serv = serv_find(parv[0])) == NULL)
		return;

	len = strlen(parv[2]);
	if (parv[2][len-1] == '.')
		--len;
	if (len > VERLEN)
		len = VERLEN;
	memcpy(serv->version, parv[2], len);
	serv->version[len] = '\0';
}

void clear_all(void)
{
	SERVPTR		server, downserv;
	DOMAINPTR	domain;
	USERPTR		user;
	LISTPTR		list, cur_ptr, next_ptr;
	int		i;

	for (i=0; i<conf.userhsize; ++i) {
		for (cur_ptr = users[i]; cur_ptr != NULL; cur_ptr = next_ptr) {
			next_ptr = cur_ptr->next;
			if ((user = cur_ptr->data) == NULL)
				continue;
			if ((user->domain != NULL)&&(user->domain->users != NULL))
				list_del_ptr(&(user->domain->users), user);
#ifndef NOBALLOC
			BlockHeapFree(list_heap, cur_ptr);
			BlockHeapFree(user_heap, user);
#else
			MyFree(cur_ptr);
			MyFree(user);
#endif
		}
		users[i] = NULL;
	}

	for (i=0; i<conf.serverhsize; ++i) {
		for (cur_ptr = servers[i]; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
			if ((server = cur_ptr->data) == NULL)
				continue;
			server->cur_users = 0;
			server->cur_opers = 0;
			server->cur_dupes = 0;
			server->links = 0;
			if (server->connected)
				server->last = now;
			server->connected = 0;
			server->uplink = NULL;
			for (list = server->servers; list != NULL; list = next_ptr) {
				next_ptr = list->next;
				downserv = list->data;
				list_del_ptr(&(server->servers), downserv);
			}
			server->servers = NULL;
			if (opts.qs) {
				list = server->users;
				while (list != NULL) {
					next_ptr = list->next;
#ifndef NOBALLOC
					BlockHeapFree(list_heap, list);
#else
					MyFree(list);
#endif
					if (next_ptr == NULL)
						break;
					list = next_ptr;
				}
			server->users = NULL;
			}
		}
	}

	for (i=0; i<conf.domainhsize; ++i) {
		for (cur_ptr = domains[i]; cur_ptr != NULL; cur_ptr = cur_ptr->next) {
			if ((domain = cur_ptr->data) == NULL)
				continue;
			if (domain->cur_users)
				domain->last = now;
			domain->cur_users = 0;
			domain->cur_dupes = 0;
			list = domain->users;
			while (list != NULL) {
				next_ptr = list->next;
#ifndef NOBALLOC
				BlockHeapFree(list_heap, list);
#else
				MyFree(list);
#endif
				if (next_ptr == NULL)
					break;
				list = next_ptr;
			}
			domain->users = NULL;
		}
	}

	totals.cur_users = 0;
	totals.cur_opers = 0;
	totals.cur_dupes = 0;
	totals.cur_servers = 1;
	net.outbuf[0] = '\0';
	net.inbuf[0] = '\0';
	net.linebuf[0] = '\0';
	net.outlen = 0;
	net.linepos = 0;
	opts.qs = 0;
}

int ignoreErrno(int ierrno)
{
	switch (ierrno) {
		case EINPROGRESS:
		case EWOULDBLOCK:
#if EAGAIN != EWOULDBLOCK
		case EAGAIN:
#endif  
		case EALREADY:
		case EINTR:
#ifdef ERESTART
		case ERESTART:
#endif
			return 1;
		default:
			return 0;
	}
}

void readln(void)
{
	int	left, num, pos, count, start=0, flags=0;
	char	*ptr;

#ifdef MSG_NOSIGNAL
	flags = MSG_NOSIGNAL;
#endif

	num = recv(pollfd[0].fd, net.inbuf, BUFSIZE, flags);
	net.inbuf[num] = '\0';
	if ((num < 1)&&(!ignoreErrno(errno))) {
		close(pollfd[0].fd);
		pollfd[0].fd = -1;
		pollfd[0].events = 0;
		connected = 0;
		clear_all();
		rotate_conn();
		return;
	}

	left = num;
	while (left > 0) {
		/* Try to find a newline where it should be (in < MAXLEN bytes) */
		count = (left < (MAXLEN-net.linepos)) ? left : (MAXLEN-net.linepos);
		if ((ptr = memccpy(net.linebuf+net.linepos, net.inbuf+start, '\n', count)) != NULL) {
			pos = ptr - (net.linebuf+net.linepos);
			if ((pos>1) && (*(ptr-2) == '\r'))
				*(ptr-2) = '\0';
			else
				*(ptr-1) = '\0';
			parse(net.linebuf);
			net.linepos = 0;
			start += pos;
			left = num-start;
		/* No newline and we're about to overflow. Parse what we can */
		} else if ((net.linepos+left) >= (MAXLEN-2)) {
			net.linebuf[MAXLEN-2-net.linepos] = '\0';
			parse(net.linebuf);
			net.linepos = 0;
			left -= (MAXLEN-2-net.linepos);
			start += (MAXLEN-2-net.linepos);
		/* No newline, but we still have room left. Just copy it all */
		} else {
			net.linepos += left;
			net.linebuf[net.linepos] = '\0';
			left = 0;
		}
	}
}

void ioloop(void)
{
	int		res;

	net.linepos=0;
	net.outlen=0;
	pollfd[0].fd = -1;
	pollfd[0].events = 0;

	for (;;) {
		if (stop) {
			check_old();
			write_data();
			if (connected) {
				writeln(":%s SQUIT %s :Shutting down", conf.servname, conf.servname);
				close(pollfd[0].fd);
			}
			exit(2);
		}
		if (!connected) {
			while (1) {
				if (server_connect())
					break;
			}
		}
		/* Delay for as long as possible, up until the next check time. */
		res = poll(pollfd, 1, ((nextc-now)*1000));
		if (res < 0) {
			if (ignoreErrno(errno))
                                continue;
			fprintf(stderr, "poll() error\n");
			exit(1);
		}
		now = time(NULL);
		if (!connected) {
			lastp = now;
			nextc = now + conf.calctime;
			continue;
		}
		if ((now - lastp) > 90) {
			lastp = now;
			writeln("PING :%s", conf.servname);
		}
		if (now >= nextc) {
			nextc = now + conf.calctime;
			check_old();
			write_data();
#ifndef NOBALLOC
			heap_garbage_collection();
#endif
		}
		if (!res)
			continue;
		if ((pollfd[0].revents & POLLHUP)||(pollfd[0].revents & POLLERR)||(pollfd[0].revents & POLLNVAL)) {
			close(pollfd[0].fd);
			pollfd[0].fd = -1;
			pollfd[0].events = 0;
			connected = 0;
			clear_all();
			rotate_conn();
			continue;
		}
		if ((pollfd[0].revents & POLLOUT)
#ifdef POLLWRNORM
		    ||(pollfd[0].revents & POLLWRNORM)
#endif
#ifdef POLLWRBAND
		    ||(pollfd[0].revents & POLLWRBAND)
#endif
		    )
			if (net.outlen)
				send_queued();
		if ((pollfd[0].revents & POLLIN)||(pollfd[0].revents & POLLPRI)
#ifdef POLLRDNORM
		    ||(pollfd[0].revents & POLLRDNORM)
#endif
#ifdef POLLRDBAND
		    ||(pollfd[0].revents & POLLRDBAND)
#endif
		    )
			readln();
	}
}

void send_queued(void)
{
	int	flags=0;

#ifdef MSG_NOSIGNAL
	flags = MSG_NOSIGNAL;
#endif

	if (send(pollfd[0].fd, net.outbuf, net.outlen, flags)<0) {
		if (!ignoreErrno(errno)) {
			close(pollfd[0].fd);
			pollfd[0].fd = -1;
			pollfd[0].events = 0;
			connected = 0;
			clear_all();
			rotate_conn();
		}
		return;
	}

	net.outbuf[0] = '\0';
	net.outlen = 0;
	pollfd[0].events = POLLIN | POLLERR;
}

void writeln(const char *pattern, ...)
{
	int	len;
	va_list	va;

	/* If we might overflow, flush the queue NOW */
	if ((net.outlen+MAXLEN) > BUFSIZE)
		send_queued();

	va_start(va, pattern);
	len=vsnprintf(net.outbuf+net.outlen, MAXLEN+1, pattern, va);
	va_end(va);
	if (len > MAXLEN-2) {
		net.outbuf[net.outlen+MAXLEN-2] = '\r';
		net.outbuf[net.outlen+MAXLEN-1] = '\n';
		net.outbuf[net.outlen+MAXLEN] = '\0';
		len = MAXLEN;
	} else {
		net.outbuf[net.outlen+len++] = '\r';
		net.outbuf[net.outlen+len++] = '\n';
		net.outbuf[net.outlen+len] = '\0';
	}
	
	if (opts.debug) {
		printf("%s", net.outbuf+net.outlen);
		fflush(stdout);
	}

	if (opts.logfile != NULL)
		write(logfd, net.outbuf+net.outlen, len);

	net.outlen += len;
	net.outbuf[net.outlen] = '\0';
	pollfd[0].events = POLLIN | POLLOUT | POLLERR;
}

int server_connect(void)
{
	struct sockaddr_in	sa;
	struct hostent		*hp;
	int			s;
	CONNPTR			conn;

	if ((conn = cur_conn->data) == NULL) {
		rotate_conn();
		return 0;
	}

	if ((hp=gethostbyname(conn->host))==NULL) {
		rotate_conn();
		return 0;
	}
	bzero(&sa, sizeof(sa));
	bcopy(hp->h_addr, &sa.sin_addr, hp->h_length);
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons((unsigned short)conn->port);

	if((s=socket(hp->h_addrtype, SOCK_STREAM, 0)) < 0) {
		rotate_conn();
		return 0;
	}
	if(connect(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
		rotate_conn();
		close(s);
		return 0;
	}

	if (opts.logfile != NULL)
		logfd = open(opts.logfile, O_WRONLY|O_CREAT|O_APPEND, S_IRUSR|S_IWUSR);
	now = time(NULL);
	pollfd[0].fd = s;
	pollfd[0].events = POLLIN | POLLERR;
	writeln("PASS %s :TS", conn->pass);
	writeln("CAPAB :QS FDF");
	writeln("SERVER %s 1 :%s", conf.servname, conf.servdesc);
	writeln("SVINFO %d %d 0 :%ld", TS_CURRENT, TS_MIN, now);
	writeln("PING :%s", conf.servname);
	lastp = now;
	connected = 1;
	return(1);
}

