#include "alcstats.h"
#include "list.h"
#include "hash.h"
#include "balloc.h"
#include "servlist.h"
#include "userlist.h"
#ifdef USE_RRD
#include "rrdtool.h"
#endif

#ifndef NOBALLOC
BlockHeap *server_heap;
#endif
LISTPTR *servers;

SERVPTR serv_add(char *server)
{
	LISTPTR		ptr;
	SERVPTR		serv;
	unsigned int	key;

	key = hash_server(server);
#ifndef NOBALLOC
	serv = BlockHeapAlloc(server_heap);
#else
	serv = MyMalloc(sizeof(SERVERS));
#endif
	ptr = list_add_ptr(&(servers[key]), serv);
	mystrlcpy(serv->servername, server, HOSTLEN);
	++totals.tot_servers;
	return serv;
}

void serv_del(char *server, unsigned int flagsplit)
{
	SERVPTR	serv, downserv;
	LISTPTR	ptr, next;
	USERPTR	user;
	
	--totals.cur_servers;
	++totals.splits;

	if ((serv = serv_find(server)) == NULL)
		return;

	for (ptr = serv->servers; ptr != NULL; ptr = next) {
		next = ptr->next;
		downserv = ptr->data;
		serv_del(downserv->servername, 0);
	}

	if (serv->uplink != NULL) {
		--serv->uplink->links;
		list_del_ptr(&(serv->uplink->servers), serv);
	}

	serv->servers = NULL;
	serv->uplink = NULL;
	serv->links = 0;
	serv->connected = 0;
	serv->last = now;
	if (flagsplit)
		++serv->splits;

	if (opts.qs) {
		for (ptr = serv->users; ptr != NULL; ptr = next) {
			next = ptr->next;
			if (ptr->data != NULL) {
				user = ptr->data;
				user_signoff(user->nickname);
			}
		}
		serv->users = NULL;
	}

	return;
}

void serv_rm(SERVPTR serv)
{
	unsigned int	key;
#ifdef USE_RRD
	deleterrd(serv->servername);
#endif
	key = hash_server(serv->servername);
	list_del_ptr(&(servers[key]), serv);
#ifndef NOBALLOC
	BlockHeapFree(server_heap, serv);
#else
	MyFree(serv);
#endif
	--totals.tot_servers;
}

SERVPTR serv_find(char *server)
{
	LISTPTR		ptr;
	SERVPTR		serv;
	unsigned int	key;

	key = hash_server(server);
	for (ptr = servers[key]; ptr != NULL; ptr = ptr->next) {
		if ((serv = ptr->data) == NULL)
			continue;
		if (!strcmp(serv->servername, server)) 
			return(serv);
	}

	return NULL;
}

void server_init(void)
{
	int	i;

#ifndef NOBALLOC
	if (!conf.serverheap)
		conf.serverheap = (int)((double)conf.serverhsize*0.95);
	server_heap = BlockHeapCreate(sizeof(SERVERS), conf.serverheap);
#endif
	servers = MyMalloc(sizeof(LISTPTR)*conf.serverhsize);
	for (i=0; i<conf.serverhsize; ++i)
		servers[i] = NULL;
}

