#include "alcstats.h"
#include "list.h"
#include "hash.h"
#include "balloc.h"
#include "servlist.h"
#include "userlist.h"
#include "domainlist.h"

#ifndef NOBALLOC
BlockHeap *user_heap;
#endif
LISTPTR *users;

USERPTR user_add(char *nickname, char *username, char *hostname, struct server_struct *server)
{
	LISTPTR		ptr;
	USERPTR		user;
	unsigned int	key;

	key = hash_nick_name(nickname);
	downstring(username);
	downstring(hostname);
#ifndef NOBALLOC
	user = BlockHeapAlloc(user_heap);
#else
	user = MyMalloc(sizeof(USERS));
#endif
	ptr = list_add_ptr(&(users[key]), user);
	mystrlcpy(user->username, username, USERLEN);
	mystrlcpy(user->hostname, hostname, HOSTLEN);
	mystrlcpy(user->nickname, nickname, NICKLEN);
	user->server = server;
	if (opts.qs)
		list_add_ptr(&(server->users), user);
	server->last = now;
	return user;
}

void user_del(char *nickname)
{
	USERPTR		user;
	DOMAINPTR	dom;
	SERVPTR		serv;
	unsigned int	key;

	key = hash_nick_name(nickname);
	if ((user = user_find(nickname)) == NULL)
		return;
	if ((user->domain != NULL)&&(user->domain->users != NULL)) {
		dom = user->domain;
		dom->last = now;
		list_del_ptr(&(user->domain->users), user);
	}
	if (opts.qs) {
		if ((user->server != NULL)&&(user->server->users != NULL)) {
			serv = user->server;
			serv->last = now;
			list_del_ptr(&(user->server->users), user);
		}
	}
	list_del_ptr(&(users[key]), user);
#ifndef NOBALLOC
	BlockHeapFree(user_heap, user);
#else
	MyFree(user);
#endif
}

USERPTR user_find(char *nickname)
{
	LISTPTR		ptr;
	USERPTR		user;
	unsigned int	key;

	key = hash_nick_name(nickname);
	for (ptr = users[key]; ptr != NULL; ptr = ptr->next) {
		if ((user = ptr->data) == NULL)
			continue;
		if (!strcmp(user->nickname, nickname))
			return(user);
	}

	return NULL;
}

/* user_signoff
 * Not really a command, but a clean way of completely removing a user
 */
void user_signoff(char *nickname)
{
	USERPTR user;

	/* I used to decrease the counters no matter what.
	 * Except there's a bug in SNICK that sends both a KILL and a QUIT. */
	if ((user = user_find(nickname)) == NULL)
		return;
	if (totals.cur_users)
		--totals.cur_users;
	/* It'd be nice to get rid of this function and to do this work in user_del(),
	 * but user_del() is also called during nick changes! *sigh* */
	if (user->server != NULL) {
		user->server->last = now;
		if (user->server->cur_users)
			--user->server->cur_users;
		if (user->oper) {
			if (totals.cur_opers)
				--totals.cur_opers;
			if (user->server->cur_opers)
				--user->server->cur_opers;
			if (user->domain != NULL)
				if (user->domain->cur_opers)
					--user->domain->cur_opers;
		}
	}
	/* I keep screwing this up, so I better comment it now that I got it right!
	 * If we're signing off a user that has dupes, we decrease the counters.
	 * If they're not a dupe themselves, however, we also unmark the first dupe
	 * of theirs that we find! */
	if (user->domain != NULL) {
		if (user->domain->cur_users)
			--user->domain->cur_users;
		if (user->domain->users != NULL)
			list_del_ptr(&(user->domain->users), user);
		if ((user->dupe)||(dupe_check(user->domain->users, user->username, user->hostname))) {
			if (totals.cur_dupes)
				--totals.cur_dupes;
			if (user->domain->cur_dupes)
				--user->domain->cur_dupes;
			if (user->dupe) {
				if (user->server->cur_dupes)
					--user->server->cur_dupes;
				user->dupe = 0;
			} else
				unmark_dupe(user->domain->users, user->username, user->hostname);
		}
		user->domain = NULL; /* So we don't try to list_del_ptr() again in user_del() */
	}
	user_del(nickname);
}

void user_init(void)
{
	int	i;

#ifndef NOBALLOC
	if (!conf.userheap)
		conf.userheap = (int)((double)conf.userhsize*0.95);
	user_heap = BlockHeapCreate(sizeof(USERS), conf.userheap);
#endif
	users = MyMalloc(sizeof(LISTPTR)*conf.userhsize);
	for (i=0; i<conf.userhsize; ++i)
		users[i] = NULL;
}

